__author__ = """Jimmy Shen"""
__email__ = "jmmshn@gmail.com"
__version__ = "0.1.0"

import os

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    pass

test_files_dir = os.path.dirname(os.path.realpath(__file__)) + "/../../test_files"



"""
Helper functions to visualize the data in plotly
"""

import numpy as np
import plotly.graph_objs as go

"""Visualizaiton functions do the scatter plots in plotly since it seems to be more efficient."""


def get_plotly_scatter_plot(
    data_in: np.ndarray,
    lat_mat: np.ndarray,
    skips: int = 5,
    logcolor: bool = False,
    mask: np.ndarray = None,
    opacity: float = 0.5,
    marker_size: int = 5,
) -> go.Figure:
    """
    Returns a plotly fig object for plotting.
    Args:
        data_in: Structured grid data to be plotted
        lat_mat: Lattice vectors of the cell
        skips: reduction factor of the grid points for plotting, only show [::skips] in each direction
        logcolor: If True, assign the color in log scale
        mask: Filter the points to plot
        opacity: opacity of each point being plotted
        marker_size: size of the markers in the 3D scatter plot

    Returns:
        plotly Figure object

    """
    ndim = len(data_in.shape)
    if ndim > 3:
        raise NotImplementedError("Can only render data of 1, 2, or 3 dimensions.")

    ss = slice(0, None, skips)
    trimmed_data = np.real(data_in).copy()
    trimmed_data = trimmed_data[(ss,) * ndim]

    if mask is not None:
        flat_mask = mask[(ss,) * ndim].flatten()
    else:
        flat_mask = np.ones_like(trimmed_data, dtype=bool).flatten()

    vecs = [
        np.linspace(0, 1, trimmed_data.shape[_], endpoint=False) for _ in range(ndim)
    ]
    gridded = np.meshgrid(*vecs, indexing="ij")  # indexing to match the labeled array
    res = np.dot(lat_mat.T, [g_.flatten() for g_ in gridded])

    if logcolor:
        cc = np.log(trimmed_data.flatten())
    else:
        cc = trimmed_data.flatten()

    if ndim == 1:
        xx = res[flat_mask]
    elif ndim > 1:
        xx = res[0, flat_mask]
        yy = res[1, flat_mask]
    if ndim > 2:
        zz = res[2, flat_mask]

    cc = cc[flat_mask]

    if ndim == 1:
        data = go.Scatter(
            x=xx,
            y=cc,
            mode="markers",
            marker=dict(
                size=marker_size,
                color="red",
            ),
        )
    if ndim == 2:
        data = go.Scatter(
            x=xx,
            y=yy,
            mode="markers",
            marker=dict(
                size=marker_size,
                color=cc,  # set color to an array/list of desired values
                colorscale="Viridis",  # choose a colorscale
                opacity=opacity,
            ),
        )
    if ndim == 3:
        data = go.Scatter3d(
            x=xx,
            y=yy,
            z=zz,
            mode="markers",
            marker=dict(
                size=marker_size,
                color=cc,
                colorscale="Viridis",
                opacity=opacity,
            ),
        )
    fig = go.Figure(data=[data])

    fig.update_layout(template="plotly_white")
    if ndim == 2:
        fig.update_layout(
            width=800, height=800, yaxis=dict(scaleanchor="x", scaleratio=1)
        )
    if ndim == 3:
        fig.update_layout(width=800, height=800, scene_aspectmode="data")
    return fig



from __future__ import annotations

from itertools import combinations
from typing import Iterable, List, Tuple, Union

import numpy as np
import numpy.typing as npt
from scipy.interpolate import RegularGridInterpolator
from scipy.ndimage import convolve


def pad_arr(arr_in: np.ndarray, shape: List[int]) -> np.ndarray:
    """
    Padding a function on a hypercube.

    Args:
        arr_in: data to be padded with zeros
        shape: desired shape of the array

    Returns:
        ndarray: padded data

    Notes:
    We basically need to move the data at the corners a hypercube to the corners of a bigger/smaller hypercube
    Each corner of the hypercube can be represented as a binary number with length equal to the dimension of the cube.
    Loop over the binary representation to figure out which corner you are at the slice accordingly.

    Examples:
    >>> arr = np.array([[1,2,3], [4,5,6]])
    >>> pad_arr(arr, [5,3])
    array([[1, 2, 3],
           [0, 0, 0],
           [0, 0, 0],
           [0, 0, 0],
           [4, 5, 6]])
    >>> pad_arr(arr, [2,2])
    array([[1, 3],
           [4, 6]])
    """

    # for _, isize in enumerate(shape):
    #     if isize < arr_in.shape[_]:
    #         raise Warning(
    #             "Some dimension of output array is smaller than the same dimension of input array."
    #         )

    def get_slice(idig, idim, bound_pairs):
        if idig == "0":
            return slice(0, bound_pairs[idim][0])
        elif idig == "1":
            return slice(-bound_pairs[idim][1], None)
        else:
            raise ValueError("Binary digit not 1 or 0")

    dimensions = arr_in.shape
    boundaries = [
        (
            int(np.ceil(min(i_dim, j_dim) + 1) / 2.0),
            int(np.floor(min(i_dim, j_dim)) / 2.0),
        )
        for i_dim, j_dim in zip(dimensions, shape)
    ]
    dim = len(dimensions)
    fmt = f"#0{dim+2}b"
    corners = [format(itr, fmt)[-dim:] for itr in range(2**dim)]
    arr_out = np.zeros(shape, dtype=arr_in.dtype)

    for ic in corners:
        islice = tuple(
            get_slice(idig, idim, boundaries) for idim, idig in enumerate(ic)
        )
        arr_out[islice] = arr_in[islice]
    return arr_out


def interpolate_fourier(arr_in: np.ndarray, shape: List[int]) -> np.ndarray:
    """
        Interpolate the data to some final shape, keep magnitude the same.
        Will perform best if the input array is periodic in all directions.

        Args:
            arr_in: Input array of data
            shape: Desired shape shape of the interpolated data

        Returns:
            interpolated data in the desired shape

        Examples:
    >>> arr = np.array([[5.0, 10.0, 10.0, 7.0],
    ...        [22.0, 12.0, 7.0, 3.0],
    ...        [16.0, 10.0, 3.0, 5.0],
    ...        [16.0, 22.0, 16.0, 6.0],
    ...        [19.0, 20.0, 3.0, 7.0]])
    >>> np.round(interpolate_fourier(arr_in=arr, shape=[10,8]),2)
    array([[ 5.  ,  7.29, 10.  , 10.83, 10.  ,  8.71,  7.  ,  5.17],
           [12.02, 12.22, 11.22, 12.72, 12.72,  8.11,  4.71,  7.62],
           [22.  , 19.49, 12.  ,  8.88,  7.  ,  2.51,  3.  , 13.12],
           [23.2 , 19.82, 10.33,  3.87,  0.64, -1.05,  3.36, 14.9 ],
           [16.  , 14.86, 10.  ,  5.67,  3.  ,  2.14,  5.  , 11.33],
           [11.58, 14.07, 14.8 , 14.24, 11.82,  8.06,  6.05,  7.88],
           [16.  , 20.66, 22.  , 20.66, 16.  ,  9.34,  6.  ,  9.34],
           [21.86, 26.35, 24.64, 18.31, 10.48,  5.16,  6.05, 13.21],
           [19.  , 22.5 , 20.  , 11.19,  3.  ,  2.  ,  7.  , 13.31],
           [ 9.34, 12.33, 13.01,  8.08,  3.33,  4.42,  7.84,  8.67]])

    """
    fft_res = np.fft.fftn(arr_in)
    fft_res = pad_arr(fft_res, shape)
    results = np.fft.ifftn(fft_res) * np.size(fft_res) / np.size(arr_in)
    # take the real value if the input array is real
    if not np.iscomplexobj(arr_in):
        return np.real(results)
    return results


def roll_array(arr: np.ndarray, roll_vec: List[int]) -> np.ndarray:
    """
    Shift the index of an ndarray based on roll_vec.
    Args:
        arr: array to be rolled
        roll_vec: number of indices in each direction to roll

    Returns:
        The rolled array
    """
    for ii, roll_val in enumerate(roll_vec):
        arr = np.roll(arr, roll_val, ii)
    return arr


def get_sc_interp(
    data_in: np.ndarray,
    sc_mat: npt.ArrayLike,
    grid_sizes: List[int],
    scipy_interp_method="linear",
    origin: Union[np.ndarray, List[float], Tuple[float]] = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Take a data array defined on a regular lattice and a new set of lattice
    vector (in units of the the original lattice), reproduce the data on the new grid.
    We can consider the original cell to be unit cell (UC) and the new cell to be a super cell (SC)
    defined using the basis of the UC -- although the the SC in this case can be any lattice.

    Args:
        data_in: data stored on a regular grid
        sc_mat: lattice vectors of new cell in the units of the old cell
        grid_sizes: number of grid points in each direction in the new cell
        scipy_interp_method: interpolation method to be used
        origin: shift applied to the origin in fractional coordinates
    Returns:
        size (ndim x prod(grid_size)) the cartesian coordinates of each point in the new data
        size (prod(grid_size)) the regridded data

    Note:
    Assuming we have some kind of data stored on a regular grid --- [0,1) in each dimension
    In the frame where UC is orthonormal we can just map all points in the sc_grid to the a position in the unit cube
    Then perform interpolation on the new coordinates that are all within the cube

    Suggestion:
        This algorithm can be used in real space and fourier space
        For real space we expect the data to be smooth, so we should use "linear" interpolation
        For fourier space we expect the position of the exact fourier component is important so use "nearest"

    We have to include the boundaries of the UC since the mapped grid must lie between the existing points
    """

    # We will need to interpolated near the boundaries so we have to pad the data
    padded_data = get_padded_array(data_in)

    # interpolate the padded data to the sc coordinate in the cube
    uc_vecs = [
        np.linspace(0, 1, isize + 1, endpoint=True) for isize in data_in.shape
    ]  # need to go from ij indexing to xy
    interp_func = RegularGridInterpolator(
        uc_vecs, padded_data, method=scipy_interp_method
    )  # input data from CHGCAR requires transpose
    grid_vec = [np.linspace(0, 1, isize, endpoint=False) for isize in grid_sizes]
    frac_coords = np.meshgrid(
        *grid_vec, indexing="ij"
    )  # indexing to match the labeled array
    frac_coords = np.vstack([icoord.flatten() for icoord in frac_coords])

    sc_coord = np.dot(np.array(sc_mat).T, frac_coords)  # shape (dim, NGRID)

    if origin is not None:
        sc_coord += np.array([[_] for _ in origin])

    mapped_coords = sc_coord - np.floor(sc_coord)

    return sc_coord, interp_func(mapped_coords.T)


def get_padded_array(data_in: np.ndarray) -> np.ndarray:
    """
    Pad an array once in each direction with the periodic boundary conditions.
    Args:
        data_in: Array to be padded

    Returns:
        Padded array

    """
    padded_data = data_in.copy()
    slice_arr = [
        [slice(0, 1) if ii == i else slice(0, None) for ii in range(len(data_in.shape))]
        for i in range(len(data_in.shape))
    ]
    for idim, islice in enumerate(slice_arr):
        padded_data = np.concatenate(
            (padded_data, padded_data[tuple(islice)]), axis=idim
        )
    return padded_data


def get_plane_spacing(lattice: np.ndarray) -> Iterable[float]:
    """
        Get the cartesian spacing between bonding planes of a unit cell
        Args:
            lattice: list of lattice vectors in cartesian coordinates

        Returns:
            List where the k-th element is is the spacing of planes generated by all
            lattice vectors EXCEPT the k-th one

        Examples:
    >>> get_plane_spacing([[1,0,0], [1,1,0], [0,0,2]]) # doctest: +ELLIPSIS
    [0.7653..., 1.042..., 2.0]

    """
    # get all pairwise projections i must be smaller than j
    ndim = len(lattice)
    idx_pairs = [*combinations(range(ndim), 2)]
    latt_len = [np.linalg.norm(lattice[i]) for i in range(ndim)]
    pproj = {
        (i, j): np.dot(lattice[i], lattice[j]) / latt_len[i] / latt_len[j]
        for i, j in idx_pairs
    }
    # get the spacing in each direction:
    spacing = []
    for idir in range(ndim):
        idir_proj = [
            np.array(lattice[j]) * pproj[tuple(sorted([idir, j]))]  # type: ignore
            for j in range(ndim)
            if j != idir
        ]
        v_perp_subspace = lattice[idir] - sum(idir_proj)
        spacing.append(np.linalg.norm(v_perp_subspace))
    return spacing


def get_ucell_frac_fit_sphere(lattice: np.ndarray, r: float = 0.2) -> Iterable[float]:
    """Lattice ffractions in each direction that are within r of the origin

    Get the smallest you can make make the lattice parameter in each direction to fit a
    sphere of radius `r`. For lattice vector `k`, the sphere must be contained within
    the hyperplanes defined by all lattice vectors except `k`.

    Parameters
    ----------
    lattice:
        List of lattice vectors in cartesian coordinates
    r:
        width of the sphere

    Returns
    -------
    Iterable of floats
        fraction of lattice vector in each direction need to fit the sphere

    Examples:
    >>> get_ucell_frac_fit_sphere([[1,0,0], [1,-1, 0], [0,0,2]], 0.1) # doctest: +ELLIPSIS
    [0.2613..., 0.1919..., 0.1]
    """
    rfrac = []
    for ispace in get_plane_spacing(lattice=lattice):
        rfrac.append(2 * r / ispace)
    return rfrac


def gaussian_smear(
    arr: npt.NDArray,
    lattice: npt.NDArray,
    sigma: float = 0.2,
    multiple: float = 4.0,
) -> Tuple[npt.NDArray, npt.NDArray]:
    """Applies an isotropic Gaussian smearing to periodic data

    Apply a Gaussian smearing of width (standard deviation) `sigma` to
    the periodic field.  The smearing obeys periodic boundary conditions at
    the edges of the cell.

    Parameters
    ----------
    arr_in:
        input data array to smear, if None: smear self.grid_data
    lattice:
        lattice vectors in cartesian coordinates
    sigma:
        Smearing width in cartesian coordinates, in the same units as the lattice vectors
    """
    # Since smearing requires floating point, we need to make sure the input is floating point
    arr = arr.astype(np.float64)

    # if the input is 1 dimensional, we need to make it 2D but and transpose it so that the
    # arr.shape[0] is the size of the first dimension
    is_1d = len(arr.shape) == 1
    if is_1d:
        arr = arr.reshape(len(arr), 1)
        lattice = lattice.reshape(1, 1)

    # get the dimension of the filter needed for to cover multiples of the smearing width
    r_frac = get_ucell_frac_fit_sphere(lattice=lattice, r=sigma * multiple)

    # we want the center of the smearing mask to be the max value so there should an odd number
    # however based on how periodic boundary conditions are implemented, we need to make sure
    # to pad boundary with an extra layer so the masks should be even
    filter_shape = [
        int(
            np.ceil(itr_rf * itr_dim / 2) * 2
        )  # size of the mask should be even in each direction to avoid edge effects
        for itr_rf, itr_dim in zip(r_frac, arr.shape)
    ]

    filter_latt = np.array(
        [
            (filter_shape[_] + 1) / (arr.shape[_] + 1) * lattice[_]
            for _ in range(len(lattice))
        ]
    )

    # Get the fractional positions
    filter_frac_c = [np.linspace(0, 1, ng, endpoint=False) for ng in filter_shape]
    frac_pos = np.meshgrid(*filter_frac_c, indexing="ij")
    frac_pos = [i_dir.flatten() for i_dir in frac_pos]

    # convert to cartesian
    cart_pos = np.matmul(filter_latt.T, np.vstack(frac_pos))

    mid_point = np.sum(filter_latt, axis=0) / 2
    disp2mid2 = [
        (i_coord.reshape(filter_shape) - mp_coord) ** 2
        for mp_coord, i_coord in zip(mid_point, cart_pos)
    ]
    dist2mid = np.sqrt(sum(disp2mid2))
    # make sure the mask is zero outside the sphere
    mm = dist2mid <= sigma * multiple
    gauss = np.exp(-1 / 2 * (dist2mid / sigma) ** 2) * mm
    gauss = gauss / gauss.sum()
    # reduce dimension before convolution if the input is 1D
    if is_1d:
        return convolve(input=arr.flatten(), weights=gauss, mode="wrap"), gauss
    return convolve(input=arr, weights=gauss, mode="wrap"), gauss



from __future__ import annotations

import warnings
from dataclasses import dataclass

"""Chang Density Objects: Periodic Grid + Lattice / Atoms"""
import math
from typing import Dict, List, Tuple, Union

import numpy as np
import numpy.typing as npt
from monty.dev import deprecated
from monty.json import MSONable
from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
from pymatgen.io.vasp import Chgcar, Poscar, VolumetricData

from pyrho.pgrid import PGrid
from pyrho.utils import get_sc_interp


@dataclass
class ChargeDensity(MSONable):
    """
    Defines a charge density with a PGrid object along with the atomic structure

    Args:
        grid_data: Volumetric data to read in
        structure: Atomic structure corresponding to the charge density
        normalization: the normalization scheme:
            - 'vasp' sum of the data / number of grid points == number of electrons
            - None/"none" no normalization
    """

    pgrids: Dict[str, PGrid]
    structure: Structure
    normalization: str | None = "vasp"

    def __post_init__(self):
        """Post initialization:
        Steps:
            - Make sure all the lattices are identical
        """
        lattices = [self.pgrids[key].lattice for key in self.pgrids.keys()]
        if not all(
            np.allclose(self.structure.lattice._matrix, lattice) for lattice in lattices
        ):
            raise ValueError("Lattices are not identical")

    @property
    def normalized_data(self) -> dict[str, npt.NDArray]:
        """Get the normalized data.

        Since different codes use different normalization methods for
        volumetric data we should convert them to the same units (electrons / Angstrom^3)

        Returns
        -------
        dict[str, NDArray]:
            The normalized data in units of (electrons / Angstrom^3)
        """
        return {
            k: _normalize_data(
                grid_data=v.grid_data,
                lattice=self.structure.lattice,
                normalization=self.normalization,
            )
            for k, v in self.pgrids.items()
        }

    @property
    def normalized_pgrids(self) -> dict[str, PGrid]:
        """Get the normalized pgrids.

        Since different codes use different normalization methods for
        volumetric data we should convert them to the same units (electrons / Angstrom^3)

        Returns
        -------
        dict[str, PGrid]:
            The normalized pgrids in units of (electrons / Angstrom^3)
        """
        return {
            k: PGrid(
                grid_data=_normalize_data(
                    grid_data=v.grid_data,
                    lattice=self.structure.lattice,
                    normalization=self.normalization,
                ),
                lattice=v.lattice,
            )
            for k, v in self.pgrids.items()
        }

    @property
    def lattice(self) -> npt.NDArray:
        """Lattice represented as an NDArray."""
        return self.structure.lattice.matrix

    @classmethod
    def from_pmg(
        cls, vdata: VolumetricData, data_key: str = "total", normalization: str = "vasp"
    ) -> "ChargeDensity":
        """
        Read a single key from the data field of a VolumetricData object
        Args:
            vdata: The volumetric data object
            data_key: The key to read from in the data field

        Returns:
            ChargeDensity object
        """
        pgrids = {
            k: PGrid(v, vdata.structure.lattice._matrix) for k, v in vdata.data.items()
        }
        return cls(
            pgrids=pgrids, structure=vdata.structure, normalization=normalization
        )

    def reorient_axis(self) -> None:
        """
        Change the orientation of the lattice vector so that:
        a points along the x-axis, b is in the xy-plane, c is in the positive-z halve of space
        """
        args: Tuple[float, float, float, float, float, float] = (
            self.structure.lattice.abc + self.structure.lattice.angles
        )
        self.structure.lattice = Lattice.from_parameters(*args, vesta=True)

    def get_data_in_cube(self, s: float, ngrid: int, key: str = "total") -> npt.NDArray:
        """Return the charge density data sampled on a cube.

        Obtain a cubic basic cubic crop of the normalized charge density data.

        Parameters
        ----------
        s:
            The side length of the cube
        ngrid:
            Number of grid points in each direction

        Returns
        -------
        NDArray:
            Regridded data in a ngrid x ngrid x ngrid array

        """
        grid_out = [ngrid, ngrid, ngrid]
        target_sc_lat_vecs = np.eye(3, 3) * s
        sc_mat = np.linalg.inv(self.structure.lattice.matrix) @ target_sc_lat_vecs
        _, res = get_sc_interp(self.normalized_data[key], sc_mat, grid_out)
        return res.reshape(grid_out)

    def get_transformed(
        self,
        sc_mat: npt.NDArray,
        grid_out: Union[List[int], int],
        origin: npt.ArrayLike = (0, 0, 0),
        up_sample: int = 1,
    ) -> "ChargeDensity":
        """Modify the structure and data and return a new object containing the reshaped data.

        Parameters
        ----------
        sc_mat:
            The transformation matrix to apply to the lattice vectors
        grid_out:
            The dimensions of the transformed grid
        origin:
            Origin of the new lattice in fractional coordinates of the input cell
        up_sample:
            The factor to scale up the sampling of the grid data using Fourier interpolation

        Returns
        -------
        ChargeDensity:
            The transformed ChargeDensity object

        """

        # warning if the sc_mat is not integer valued
        if not np.allclose(np.round(sc_mat), sc_mat):
            warnings.warn(
                "The `sc_mat` is not integer valued.\n"
                "Non-integer valued transformations are valid but will not product periodic structures, thus we cannot define a new Structure object.\n"
                "We will round the sc_mat to integer values for now but can implement functionality that returns a Molecule object in the future.",
            )
        sc_mat = np.round(sc_mat).astype(int)
        new_structure = self.structure.copy()
        new_structure.translate_sites(
            list(range(len(new_structure))), -np.array(origin)
        )
        new_structure = new_structure * sc_mat

        # determine the output grid
        lengths = new_structure.lattice.abc
        if isinstance(grid_out, int):
            ngrid = grid_out / new_structure.volume
            mult = (np.prod(lengths) / ngrid) ** (1 / 3)
            grid_out = [int(math.floor(max(l_ / mult, 1))) for l_ in lengths]
        else:
            grid_out = grid_out

        pgrids = {}
        for k, pgrid in self.normalized_pgrids.items():
            new_pgrid = pgrid.get_transformed(
                sc_mat=sc_mat, grid_out=grid_out, origin=origin, up_sample=up_sample
            )
            pgrids[k] = _scaled_data(
                grid_data=new_pgrid,
                lattice=new_structure.lattice,
                normalization=self.normalization,
            )

        return ChargeDensity(
            pgrids=pgrids,
            structure=new_structure,
            normalization=self.normalization,
        )

    def to_Chgcar(self) -> Chgcar:
        """Convert the charge density to a pymatgen.io.vasp.outputs.Chgcar object

        Scale and convert each key in the pgrids dictionary and create a Chgcar object

        Returns:
            Chgcar: The charge density object
        """
        struct = self.structure.copy()
        data_dict = {}
        for k, v in self.pgrids.items():
            data_dict[k] = self._scale_data(v, normalization="vasp")
        return Chgcar(Poscar(struct), data=data_dict)

    @classmethod
    def from_file(
        cls, filename: str, pmg_obj: VolumetricData = Chgcar
    ) -> "ChargeDensity":
        """Read a ChargeDensity object from a file.

        Parameters
        ----------
        filename:
            The filename of the ChargeDensity object
        pmg_obj:
            The pymatgen object to read from the file (default: Chgcar).
            the `from_file` method from this class will be called to read the file.

        Returns
        -------
            ChargeDensity: The ChargeDensity object
        """
        return cls.from_pmg(pmg_obj.from_file(filename))

    #
    #     _, new_rho = get_sc_interp(self.rho, sc_mat, grid_sizes=grid_out)
    #     new_rho = new_rho.reshape(grid_out)
    #
    #     grid_shifts = [
    #         int(t * g) for t, g in zip(translation - np.round(translation), grid_out)
    #     ]
    #
    #     new_rho = roll_array(new_rho, grid_shifts)
    #     return self.__class__.from_rho(new_rho, new_structure)


# class SpinChargeDensity(MSONable, ChargeABC):
#     def __init__(self, chargeden_dict: Dict, aug_charge: Dict = None):
#         """
#         Wrapper class that parses multiple sets of grid data on the same lattice

#         Args:
#             chargeden_dict: A dictionary containing multiple charge density objects
#                         typically in the format {'total' : ChargeDen1, 'diff' : ChargeDen2}
#         """
#         self.chargeden_dict = chargeden_dict
#         self.aug_charge = aug_charge
#         self._tmp_key = next(
#             iter(self.chargeden_dict)
#         )  # get one key in the dictionary to make writing the subsequent code easier

#     @classmethod
#     def from_pmg_volumetric_data(
#         cls, vdata: VolumetricData, data_keys=("total", "diff")
#     ):
#         chargeden_dict = {}
#         data_aug = getattr(vdata, "data_aug", None)
#         for k in data_keys:
#             chargeden_dict[k] = ChargeDensity.from_pmg(vdata, data_key=k)
#         return cls(chargeden_dict, aug_charge=data_aug)

#     @property
#     def lattice(self) -> Lattice:
#         return self.chargeden_dict[self._tmp_key].lattice

#     def to_Chgcar(self) -> Chgcar:
#         struct = self.chargeden_dict[self._tmp_key].structure
#         data_ = {k: v.renormalized_data for k, v in self.chargeden_dict.items()}
#         return Chgcar(Poscar(struct), data_, data_aug=self.aug_charge)

#     def to_VolumetricData(self) -> VolumetricData:
#         key_ = next(iter(self.chargeden_dict))
#         struct = self.chargeden_dict[key_].structure
#         data_ = {k: v.renormalized_data for k, v in self.chargeden_dict.items()}
#         return VolumetricData(struct, data_)

#     def get_reshaped(
#         self,
#         sc_mat: npt.ArrayLike,
#         grid_out: Union[List, int],
#         origin: npt.ArrayLike = (0.0, 0.0, 0.0),
#         up_sample: int = 1,
#     ) -> "SpinChargeDensity":
#         new_spin_charge = {}
#         for k, v in self.chargeden_dict.items():
#             new_spin_charge[k] = v.get_reshaped_cell(sc_mat, frac_shift, grid_out)
#         factor = int(
#             new_spin_charge[self._tmp_key].structure.num_sites
#             / self.chargeden_dict[self._tmp_key].structure.num_sites
#         )
#         new_aug = {}
#         if self.aug_charge is not None:
#             for k, v in self.aug_charge.items():
#                 new_aug[k] = multiply_aug(v, factor)
#         return self.__class__(new_spin_charge, new_aug)

#     def reorient_axis(self) -> None:
#         for k, v in self.chargeden_dict:
#             v.reorient_axis()


@deprecated
def multiply_aug(data_aug: List[str], factor: int) -> List[str]:
    """
    The original idea here was to use to to speed up some vasp calculations for
    supercells by initializing the entire CHGCAR file.
    The current code does not deal with transformation of the Augementation charges after regridding.

    This is a naive way to multiply the Augmentation data in the CHGCAR,
    a real working implementation will require analysis of the PAW projection operators.
    However, even with such an implementation, the speed up will be minimal due to VASP's internal
    minimization algorithms.
    Args:
        data_aug: The original augmentation data from a CHGCAR
        factor: The multiplication factor (some integer number of times it gets repeated)
    Returns:
        List of strings for each line of the Augmentation data.
    """
    res: List[str] = []
    cur_block: List[str] = []
    cnt = 0
    for ll in data_aug:
        if "augmentation" in ll:
            if cur_block:
                for _ in range(factor):
                    cnt += 1
                    cur_block[
                        0
                    ] = f"augmentation occupancies{cnt:>4}{cur_block[0].split()[-1]:>4}\n"
                    res.extend(cur_block)
            cur_block = [ll]
        else:
            cur_block.append(ll)
    else:
        for _ in range(factor):
            cnt += 1
            cur_block[
                0
            ] = f"augmentation occupancies{cnt:>4}{cur_block[0].split()[-1]:>4}\n"
            res.extend(cur_block)
    return res


def _normalize_data(
    grid_data: npt.NDArray, lattice: Lattice, normalization: str | None = "vasp"
) -> npt.NDArray:
    """Normalize the data to the number of electrons

    Since different codes use different normalization methods for
    volumetric data we should convert them to the same units (electrons / Angstrom^3)

    Parameters
    ----------
    grid_data:
        The grid data to normalize
    lattice:
        The lattice that the grid data is represented on
    normalization:
        The normalization method defaults to vasp
        - None: no normalization
        - vasp:
            The standard charge density from VASP is given as (rho*V) such that:
                sum(rho)/NGRID = NELECT/vol
            so the real normalized rho should be:
                rho = (rho*vol)*NGRID/vol/vol
            where the second `/vol` account for the different number of electrons in
            different cells

    Returns
    -------
    NDArray:
        The normalized grid data
    """
    if normalization is None or normalization[0].lower() == "n":
        return grid_data
    elif normalization[0].lower() == "v":
        return grid_data / lattice.volume
    else:
        raise NotImplementedError("Not a valid normalization scheme")


def _scaled_data(
    grid_data: npt.NDArray, lattice: Lattice, normalization: str | None = "vasp"
) -> npt.NDArray:
    """Undo the normalization of the data

    Parameters
    ----------
    grid_data:
        The grid data to unnormalize
    lattice:
        The lattice that the grid data is represented on
    normalization:
        The normalization method defaults to vasp

    Returns
    -------
    NDArray:
        The unnormalized grid data
    """
    if normalization is None or normalization[0].lower() == "n":
        return grid_data
    elif normalization[0].lower() == "v":
        return grid_data * lattice.volume
    else:
        raise NotImplementedError("Not a valid normalization scheme")



"""Python class for ND grid data volumetric data."""
from __future__ import annotations

from typing import List, Union

import numpy as np
import numpy.typing as npt
from monty.json import MSONable

from pyrho.utils import gaussian_smear, get_sc_interp, interpolate_fourier


class PGrid(MSONable):
    def __init__(self, grid_data: npt.NDArray, lattice: npt.NDArray):
        """Base class for N-dimensional Regular period grid data.

        The core code is valid for any N-dimensional periodic data

        Parameters
        ----------
        grid_data:
            Data stored on the regular rid
        lattice:
            Lattice vectors of the grid
        """
        self.grid_data = grid_data
        self.lattice = np.array(lattice)
        self._dim = len(self.grid_data.shape)
        self.grid_shape = self.grid_data.shape
        self.ngridpts = np.prod(self.grid_shape)

    def _transform_data(
        self,
        sc_mat: npt.ArrayLike,
        grid_out: List[int],
        origin: npt.ArrayLike = (0, 0, 0),
        up_sample: int = 1,
    ) -> npt.NDArray:
        """Apply a supercell transformation to the grid data

        This function assumes that the data is fixed in place and the transformation
        is applied to the lattice vectors.

        Parameters
        ----------
        sc_mat:
            Matrix transformation applied to the lattice vectors
        grid_out:
            The dimensions of the output grid
        origin:
            Origin of the new lattice in fractional coordinates of the input cell
        up_sample:
            The factor to scale up the sampling of the grid data using Fourier interpolation

        Returns
        -------
        NDArray:
            The transformed data

        """
        if up_sample == 1:
            interp_grid_data = self.grid_data
        else:
            interp_grid_data = interpolate_fourier(
                arr_in=self.grid_data,
                shape=[g_dim_ * up_sample for g_dim_ in self.grid_data.shape],
            )
        _, new_data = get_sc_interp(interp_grid_data, sc_mat, grid_sizes=grid_out, origin=origin)  # type: ignore
        new_data = new_data.reshape(grid_out)
        return new_data

    def __mul__(self, factor: float) -> PGrid:
        """Multiply the grid data by a factor

        Parameters
        ----------
        factor:
            The factor to multiply the grid data by

        Returns
        -------
        PGrid:
            The new PGrid object

        """
        return PGrid(grid_data=self.grid_data * factor, lattice=self.lattice)

    def __truediv__(self, factor: float) -> PGrid:
        """Divide the grid data by a factor

        Parameters
        ----------
        factor:
            The factor to divide the grid data by

        Returns
        -------
        PGrid:
            The new PGrid object

        """
        return PGrid(grid_data=self.grid_data / factor, lattice=self.lattice)

    def get_transformed(
        self,
        sc_mat: Union[List[List[int]], npt.NDArray],
        grid_out: List[int],
        origin: npt.NDArray = (0, 0, 0),
        up_sample: int = 1,
    ) -> PGrid:
        """Get a new PGrid object for the new transformed data

        Parameters
        ----------
        sc_mat:
            Matrix transformation applied to the lattice vectors
        grid_out:
            The dimensions of the output grid
        origin:
            Origin of the new lattice in fractional coordinates of the input cell
        up_sample:
            The factor to scale up the sampling of the grid data using Fourier interpolation

        Returns
        -------
        PGrid:
            The transformed PGrid object

        """
        new_data = self._transform_data(
            sc_mat=sc_mat, grid_out=grid_out, origin=origin, up_sample=up_sample
        )
        new_lattice = np.dot(sc_mat, self.lattice)
        return PGrid(grid_data=new_data, lattice=new_lattice)

    def lossy_smooth_compression(
        self, grid_out: List, smear_std: float = 0.2
    ) -> npt.NDArray:
        """
        Perform Fourier interpolation then Gaussian smoothing.
        the smoothing makes sure that simple operation like max and min filters still
        give the same results.

        Args:
            grid_out: desired output grid of the compressed data,

        Returns:
            Smoothed array
        """
        arr_interp = np.absolute(interpolate_fourier(self.grid_data, grid_out))
        if smear_std > 0:
            arr_interp, _ = gaussian_smear(
                arr=arr_interp, lattice=self.lattice, sigma=smear_std
            )
        return arr_interp
